@@ -3,6 +3,7 @@ import collections
 import datetime
 import json
 import re
+import socket
 import string
 import tempfile
 import types
@@ -16,6 +17,7 @@ from yarl import URL
 
 from . import hdrs, multipart
 from .helpers import HeadersMixin, SimpleCookie, reify, sentinel
+from .streams import EmptyStreamReader
 from .web_exceptions import HTTPRequestEntityTooLarge
 
 
@@ -27,7 +29,7 @@ FileField = collections.namedtuple(
 _TCHAR = string.digits + string.ascii_letters + r"!#$%&'*+.^_`|~-"
 # '-' at the end to prevent interpretation as range in a char class
 
-_TOKEN = r'[{tchar}]*'.format(tchar=_TCHAR)
+_TOKEN = r'[{tchar}]+'.format(tchar=_TCHAR)
 
 _QDTEXT = r'[{}]'.format(
     r''.join(chr(c) for c in (0x09, 0x20, 0x21) + tuple(range(0x23, 0x7F))))
@@ -39,12 +41,8 @@ _QUOTED_PAIR = r'\\[\t !-~]'
 _QUOTED_STRING = r'"(?:{quoted_pair}|{qdtext})*"'.format(
     qdtext=_QDTEXT, quoted_pair=_QUOTED_PAIR)
 
-_FORWARDED_PARAMS = (
-    r'[bB][yY]|[fF][oO][rR]|[hH][oO][sS][tT]|[pP][rR][oO][tT][oO]')
-
 _FORWARDED_PAIR = (
-    r'^({forwarded_params})=({token}|{quoted_string})$'.format(
-        forwarded_params=_FORWARDED_PARAMS,
+    r'({token})=({token}|{quoted_string})'.format(
         token=_TOKEN,
         quoted_string=_QUOTED_STRING))
 
@@ -63,8 +61,13 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
     POST_METHODS = {hdrs.METH_PATCH, hdrs.METH_POST, hdrs.METH_PUT,
                     hdrs.METH_TRACE, hdrs.METH_DELETE}
 
-    def __init__(self, message, payload, protocol, writer, time_service, task,
-                 *, secure_proxy_ssl_header=None, client_max_size=1024**2):
+    def __init__(self, message, payload, protocol, writer, task,
+                 loop,
+                 *, client_max_size=1024**2,
+                 state=None,
+                 scheme=None, host=None, remote=None):
+        if state is None:
+            state = {}
         self._message = message
         self._protocol = protocol
         self._transport = protocol.transport
@@ -78,15 +81,19 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
         self._post = None
         self._read_bytes = None
 
-        self._secure_proxy_ssl_header = secure_proxy_ssl_header
-        self._time_service = time_service
-        self._state = {}
+        self._state = state
         self._cache = {}
         self._task = task
         self._client_max_size = client_max_size
+        self._loop = loop
+
+        self._scheme = scheme
+        self._host = host
+        self._remote = remote
 
     def clone(self, *, method=sentinel, rel_url=sentinel,
-              headers=sentinel):
+              headers=sentinel, scheme=sentinel, host=sentinel,
+              remote=sentinel):
         """Clone itself with replacement some attributes.
 
         Creates and returns a new instance of Request object. If no parameters
@@ -113,14 +120,24 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
         message = self._message._replace(**dct)
 
+        kwargs = {}
+        if scheme is not sentinel:
+            kwargs['scheme'] = scheme
+        if host is not sentinel:
+            kwargs['host'] = host
+        if remote is not sentinel:
+            kwargs['remote'] = remote
+
         return self.__class__(
             message,
             self._payload,
             self._protocol,
             self._writer,
-            self._time_service,
             self._task,
-            secure_proxy_ssl_header=self._secure_proxy_ssl_header)
+            self._loop,
+            client_max_size=self._client_max_size,
+            state=self._state.copy(),
+            **kwargs)
 
     @property
     def task(self):
@@ -146,6 +163,10 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
     def rel_url(self):
         return self._rel_url
 
+    @property
+    def loop(self):
+        return self._loop
+
     # MutableMapping API
 
     def __getitem__(self, key):
@@ -165,25 +186,14 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
     ########
 
-    @property
-    def scheme(self):
-        """A string representing the scheme of the request.
-
-        'http' or 'https'.
-        """
-        return self.url.scheme
-
     @property
     def secure(self):
-        """A bool indicating if the request is handled with SSL or
-        'secure_proxy_ssl_header' is matching
-
-        """
-        return self.url.scheme == 'https'
+        """A bool indicating if the request is handled with SSL."""
+        return self.scheme == 'https'
 
     @reify
     def forwarded(self):
-        """ A tuple containing all parsed Forwarded header(s).
+        """A tuple containing all parsed Forwarded header(s).
 
         Makes an effort to parse Forwarded headers as specified by RFC 7239:
 
@@ -201,54 +211,59 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
         Returns a tuple containing one or more immutable dicts
         """
-        def _parse_forwarded(forwarded_headers):
-            for field_value in forwarded_headers:
-                # by=...;for=..., For=..., BY=...
-                for forwarded_elm in field_value.split(','):
-                    # by=...;for=...
-                    fvparams = dict()
-                    forwarded_pairs = (
-                        _FORWARDED_PAIR_RE.findall(pair)
-                        for pair in forwarded_elm.strip().split(';'))
-                    for forwarded_pair in forwarded_pairs:
-                        # by=...
-                        if len(forwarded_pair) != 1:
-                            # non-compliant syntax
-                            break
-                        param, value = forwarded_pair[0]
-                        if param.lower() in fvparams:
-                            # duplicate param in field-value
-                            break
-                        if value and value[0] == '"':
-                            # quoted string: replace quotes and escape
-                            # sequences
-                            value = _QUOTED_PAIR_REPLACE_RE.sub(
-                                r'\1', value[1:-1])
-                        fvparams[param.lower()] = value
+        elems = []
+        for field_value in self._message.headers.getall(hdrs.FORWARDED, ()):
+            length = len(field_value)
+            pos = 0
+            need_separator = False
+            elem = {}
+            elems.append(types.MappingProxyType(elem))
+            while 0 <= pos < length:
+                match = _FORWARDED_PAIR_RE.match(field_value, pos)
+                if match is not None:           # got a valid forwarded-pair
+                    if need_separator:
+                        # bad syntax here, skip to next comma
+                        pos = field_value.find(',', pos)
                     else:
-                        yield types.MappingProxyType(fvparams)
-                        continue
-                    yield dict()
-
-        return tuple(
-            _parse_forwarded(self._message.headers.getall(hdrs.FORWARDED, ())))
+                        (name, value) = match.groups()
+                        if value[0] == '"':
+                            # quoted string: remove quotes and unescape
+                            value = _QUOTED_PAIR_REPLACE_RE.sub(r'\1',
+                                                                value[1:-1])
+                        elem[name.lower()] = value
+                        pos += len(match.group(0))
+                        need_separator = True
+                elif field_value[pos] == ',':      # next forwarded-element
+                    need_separator = False
+                    elem = {}
+                    elems.append(types.MappingProxyType(elem))
+                    pos += 1
+                elif field_value[pos] == ';':      # next forwarded-pair
+                    need_separator = False
+                    pos += 1
+                elif field_value[pos] in ' \t':
+                    # Allow whitespace even between forwarded-pairs, though
+                    # RFC 7239 doesn't. This simplifies code and is in line
+                    # with Postel's law.
+                    pos += 1
+                else:
+                    # bad syntax here, skip to next comma
+                    pos = field_value.find(',', pos)
+        return tuple(elems)
 
     @reify
-    def _scheme(self):
-        proto = None
+    def scheme(self):
+        """A string representing the scheme of the request.
+
+        'http' or 'https'.
+        """
+        scheme = self._scheme
+        if scheme is not None:
+            return scheme
         if self._transport.get_extra_info('sslcontext'):
-            proto = 'https'
-        elif self._secure_proxy_ssl_header is not None:
-            header, value = self._secure_proxy_ssl_header
-            if self.headers.get(header) == value:
-                proto = 'https'
+            return 'https'
         else:
-            proto = next(
-                (f['proto'] for f in self.forwarded if 'proto' in f), None
-            )
-            if not proto and hdrs.X_FORWARDED_PROTO in self._message.headers:
-                proto = self._message.headers[hdrs.X_FORWARDED_PROTO]
-        return proto or 'http'
+            return 'http'
 
     @property
     def method(self):
@@ -268,7 +283,7 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
     @reify
     def host(self):
-        """ Hostname of the request.
+        """Hostname of the request.
 
         Hostname is resolved through the following headers, in this order:
 
@@ -278,20 +293,39 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
         Returns str, or None if no hostname is found in the headers.
         """
-        host = next(
-            (f['host'] for f in self.forwarded if 'host' in f), None
-        )
-        if not host and hdrs.X_FORWARDED_HOST in self._message.headers:
-            host = self._message.headers[hdrs.X_FORWARDED_HOST]
-        elif hdrs.HOST in self._message.headers:
-            host = self._message.headers[hdrs.HOST]
-        return host
+        host = self._host
+        if host is not None:
+            return host
+        host = self._message.headers.get(hdrs.HOST)
+        if host is not None:
+            return host
+        else:
+            return socket.getfqdn()
+
+    @reify
+    def remote(self):
+        """Remote IP of client initiated HTTP request.
+
+        The IP is resolved through the following headers, in this order:
+
+        - Forwarded
+        - X-Forwarded-For
+        - peername of opened socket
+        """
+        remote = self._remote
+        if remote is not None:
+            return remote
+        transport = self._transport
+        peername = transport.get_extra_info('peername')
+        if isinstance(peername, (list, tuple)):
+            return peername[0]
+        else:
+            return peername
 
     @reify
     def url(self):
-        return URL('{}://{}{}'.format(self._scheme,
-                                      self.host,
-                                      str(self._rel_url)))
+        url = URL.build(scheme=self.scheme, host=self.host)
+        return url.join(self._rel_url)
 
     @property
     def path(self):
@@ -367,11 +401,6 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
         """Is keepalive enabled by client?"""
         return not self._message.should_close
 
-    @property
-    def time_service(self):
-        """Time service"""
-        return self._time_service
-
     @reify
     def cookies(self):
         """Return request cookies.
@@ -424,9 +453,22 @@ class BaseRequest(collections.MutableMapping, HeadersMixin):
 
     @property
     def has_body(self):
-        """Return True if request has HTTP BODY, False otherwise."""
+        """Return True if request's HTTP BODY can be read, False otherwise."""
+        warnings.warn(
+            "Deprecated, use .can_read_body #2005",
+            DeprecationWarning, stacklevel=2)
         return not self._payload.at_eof()
 
+    @property
+    def can_read_body(self):
+        """Return True if request's HTTP BODY can be read, False otherwise."""
+        return not self._payload.at_eof()
+
+    @property
+    def body_exists(self):
+        """Return True if request has HTTP BODY, False otherwise."""
+        return type(self._payload) is not EmptyStreamReader
+
     @asyncio.coroutine
     def release(self):
         """Release request.
@@ -564,6 +606,18 @@ class Request(BaseRequest):
         # or information about traversal lookup
         self._match_info = None  # initialized after route resolving
 
+    def clone(self, *, method=sentinel, rel_url=sentinel,
+              headers=sentinel, scheme=sentinel, host=sentinel,
+              remote=sentinel):
+        ret = super().clone(method=method,
+                            rel_url=rel_url,
+                            headers=headers,
+                            scheme=scheme,
+                            host=host,
+                            remote=remote)
+        ret._match_info = self._match_info
+        return ret
+
     @property
     def match_info(self):
         """Result of route resolving."""
